/*
Code to display the Minecraft portalgfx on a Nokia E66 LCD.
(C) 2011 Jeroen Domburg (jeroen AT spritesmods.com)

This program is free software: you can redistribute it and/or modify
t under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
	    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
			    
You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <libopencm3/stm32/rcc.h>
#include <libopencm3/stm32/flash.h>
#include <libopencm3/stm32/timer.h>
#include <libopencm3/stm32/gpio.h>
#include "tiles.h"
#include "particles.h"
#include <stdlib.h>


//stupid libc-ish naive random function. Good 'nuf to randomize some particles tho.
static unsigned int seed=12345;
unsigned int myrand() {
	seed = seed * 1103515245 + 12345;
	return (seed&0x7FFFFFFF);
}

void clock_setup(void) {
	// Enable internal high-speed oscillator.
	rcc_osc_on(HSI);
	rcc_wait_for_osc_ready(HSI);
	// Select HSI as SYSCLK source.
	rcc_set_sysclk_source(RCC_CFGR_SW_SYSCLKSEL_HSICLK);
	// Set prescalers for AHB, ADC, ABP1, ABP2.
	rcc_set_hpre(RCC_CFGR_HPRE_SYSCLK_NODIV);	/* Max. 72MHz */
	rcc_set_adcpre(RCC_CFGR_ADCPRE_PCLK2_DIV4);	/* Max. 14MHz */
	rcc_set_ppre1(RCC_CFGR_PPRE1_HCLK_NODIV);	/* Max. 36MHz */
	rcc_set_ppre2(RCC_CFGR_PPRE2_HCLK_NODIV);	/* Max. 72MHz */
	// Sysclk is running with 32MHz -> 1ws
	flash_set_ws(FLASH_LATENCY_1WS);
	//Set the PLL multiplication factor to 8.
	//8MHz (internal) * 8 (multiplier) / 2 (PLLSRC_HSI_CLK_DIV2) = 32MHz
	rcc_set_pll_multiplication_factor(RCC_CFGR_PLLMUL_PLL_CLK_MUL8);
	/* Select HSI/2 as PLL source. */
	rcc_set_pll_source(RCC_CFGR_PLLSRC_HSI_CLK_DIV2);
	/* Enable PLL oscillator and wait for it to stabilize. */
	rcc_osc_on(PLL);
	rcc_wait_for_osc_ready(PLL);
	/* Select PLL as SYSCLK source. */
	rcc_set_sysclk_source(RCC_CFGR_SW_SYSCLKSEL_PLLCLK);

	// Enable timer2
	rcc_peripheral_enable_clock(&RCC_APB1ENR, RCC_APB1ENR_TIM2EN);
	//Enable periph clocks
	rcc_peripheral_enable_clock(&RCC_APB2ENR, RCC_APB2ENR_IOPCEN|RCC_APB2ENR_IOPBEN|
			RCC_APB2ENR_IOPAEN|RCC_APB2ENR_AFIOEN);
}

#define LCD_BUS		0xFF
#define LCD_DNC		(1<<10)
#define LCD_NWR		(1<<11)
#define LCD_NRST	(1<<12)
#define LCD_NCS		(1<<13)
#define LCD_NRD		(1<<14)

void gpio_setup(void) {
	//Set up timer 2, pc4 for boost-converter for led
	TIM2_CR1 = TIM_CR1_CKD_CK_INT | TIM_CR1_CMS_EDGE;
	TIM2_CR2=0;
	TIM2_ARR = 256;
	TIM2_SMCR=0;
	TIM2_PSC = 0;
	TIM2_CCMR1=0;
	TIM2_CCMR2 = TIM_CCMR2_OC4M_PWM1 | TIM_CCMR2_OC4PE;
	TIM2_CCER = TIM_CCER_CC4P | TIM_CCER_CC4E;
	TIM2_CCR4 = 256-40;
	TIM2_CR1 = TIM_CR1_ARPE|TIM_CR1_CEN;
	TIM2_EGR = TIM_EGR_UG;

	AFIO_MAPR=0x4000000; //disable swj
	gpio_set_mode(GPIOA, GPIO_MODE_OUTPUT_50_MHZ, GPIO_CNF_OUTPUT_ALTFN_PUSHPULL,
			GPIO_TIM2_CH4);

	gpio_set_mode(GPIOB, GPIO_MODE_OUTPUT_50_MHZ, GPIO_CNF_OUTPUT_PUSHPULL, 
			LCD_BUS|LCD_DNC|LCD_NWR|LCD_NRD|LCD_NRST|LCD_NCS);

}

void lcdReset() {
	int i;
	GPIOB_ODR=LCD_DNC|LCD_NWR|LCD_NRD|LCD_NWR;
	for (i = 0; i < 80000; i++) __asm__("nop");
	GPIOB_ODR=LCD_DNC|LCD_NWR|LCD_NRD|LCD_NWR|LCD_NRST;
	for (i = 0; i < 80000; i++) __asm__("nop");
}

//Send a byte to the LCD. isData=0 sends a command, otherwise a data byte.
inline void lcdSendByte(int data, int isData) {
	int i,j;
	i=data&0xff;
	if (isData) i|=LCD_DNC|LCD_NRST|LCD_NRD; else i|=LCD_NRST|LCD_NRD;
	GPIOB_ODR=i;
	GPIOB_ODR=i|LCD_NWR;
}

//Data for one particle.
typedef struct {
	unsigned int image;
	int size;
	int x;
	int y;
	int dx;
	int dy;
} particleDataStruct;

//Amount of particles floating around
#define NO_PARTS 32
//Color of the particles
#define PARTCOL 0xE04AF9

//We use a fixed point (24.8) scheme to store some values, to ensure smooth 
//movement.
#define FP_POINTPOS 8
#define I2FP(x) ((x)<<FP_POINTPOS)
#define FP2I(x) ((x)>>FP_POINTPOS)


//Generates a new particle.
void newPart(particleDataStruct *part) {
	part->x=myrand()%I2FP(240);
	part->y=myrand()%I2FP(160)+I2FP(160);
	part->dx=(myrand()%I2FP(3))-I2FP(1);
	part->dy=(myrand()%I2FP(4))-I2FP(5);
	//Make sure particle moves to center
	if (part->x>I2FP(240/2) && part->dx>0) part->dx=-part->dx;
	if (part->x<I2FP(240/2) && part->dx<0) part->dx=-part->dx;
	part->image=myrand()%8;
	part->size=(myrand()%50)+10;
}

//Moves a particle one tick
void tickPart(particleDataStruct *part) {
	part->x+=part->dx;
	part->y+=part->dy;
	if (part->dy<I2FP(-1)) part->dy+=(part->y/1500);
	part->size--;
	if (part->x<I2FP(0) || part->y<I2FP(0) || 
	 part->x>I2FP(240) || part->y>I2FP(320) ||
	 part->size<0) {
		newPart(part);
	}
}

int main(void) {
	int x,y,z,tick=0;
	int tx,ty,c,n,v;
	unsigned int lineDat[240];
	unsigned int *p;
	particleDataStruct particle[NO_PARTS];

	clock_setup();
	gpio_setup();

	lcdReset();
	//Power on the LCD
	lcdSendByte(0x1, 0);
	lcdSendByte(0x11, 0);
	lcdSendByte(0x13, 0);
	lcdSendByte(0x29, 0);

	//Generate particles
	for (x=0; x<NO_PARTS; x++) {
		newPart(&particle[x]);
	}

	while(1) {
		//Move the particles before drawing the frame.
		for (x=0; x<NO_PARTS; x++) {
			tickPart(&particle[x]);
		}

		lcdSendByte(0x2C, 0); //start drawing a new frame
		for (y=0; y<320; y++) {

			//We need to draw a line. Compose it first in the linedat variable.
			//Copy the background tiles of this line to linedat.
			ty=((320-y)/8)&15;
			p=&lineDat[0];
			for (x=0; x<240; x+=8) {
				tx=(x/8)&15;
				c=portalTileGfx[tx+(ty<<4)+(tick<<8)];
				for (z=0; z<8; z++) {
					*p=c;
					p++;
				}
			}

			//Now see if we need to copy particles over that.
			for (n=0; n<NO_PARTS; n++) {
				if (particle[n].y>=I2FP(y) && particle[n].y<I2FP(y+32)) {
					//Particle needs to be drawn. Draw it!
					//particleGfx contains a bunch of 32bit bitmaps indicating if a particle
					//pixel has to be drawn or not. (=1bit image)
					v=particleGfx[(particle[n].image*8*32)+((particle[n].size>>3)*32)+(FP2I(particle[n].y)-y)];
					for (x=0; x<32; x++) {
						tx=x+FP2I(particle[n].x);
						if (v&1 && tx>=0 && tx<240) lineDat[x+FP2I(particle[n].x)]=PARTCOL;
						v>>=1;
					}
				}
			}

			//Compositing the line is done. Send to the LCD!
			p=&lineDat[0];
			for (x=0; x<240; x++) {
				lcdSendByte(*p>>16, 1);
				lcdSendByte(*p>>8, 1);
				lcdSendByte(*p, 1);
				p++;
			}
		}
		//Next frame!
		tick++;
		tick&=31;
	}

	return 0;
}
